package com.winhc.bigdata.spark.jobs.dynamic

import java.util.Date

import com.winhc.bigdata.spark.config.EsConfig
import com.winhc.bigdata.spark.utils.BaseUtil.isWindows
import com.winhc.bigdata.spark.utils.ReflectUtils.getClazz
import com.winhc.bigdata.spark.utils.{LoggingUtils, SparkUtils}
import org.apache.commons.lang3.time.DateFormatUtils
import org.apache.spark.internal.Logging
import org.apache.spark.sql.types.StringType
import org.apache.spark.sql.{Row, SparkSession}

import scala.annotation.meta.getter
import scala.collection.immutable.ListMap
import scala.collection.mutable

/**
 * @Author: XuJiakai
 * @Date: 2020/7/27 16:52
 * @Description: 企业动态
 */
object CompanyDynamic {

  case class CompanyDynamicUtil(s: SparkSession,
                                project: String, //表所在工程名
                                ds: String //此维度主键

                               ) extends LoggingUtils with Logging {
    @(transient@getter) val spark: SparkSession = s

    private val env = "dev"
    val targetTab = "xjk_tmp_company_dynamic"

    def init(): Unit = {
      sql(
        s"""
           |CREATE TABLE IF NOT EXISTS ${getEnvProjectName(env, project)}.$targetTab
           |(
           |    cid  STRING COMMENT '公司id'
           |    ,cname  STRING COMMENT '公司name'
           |    ,info_type STRING COMMENT '变更分类，大类'
           |    ,rta_desc STRING COMMENT '变更信息描述，变更标题'
           |    ,change_content STRING COMMENT '变更内容'
           |    ,change_time STRING COMMENT '变更时间'
           |    ,biz_id STRING COMMENT '业务id，数据行id'
           |    ,sub_info_type STRING COMMENT '变更小类，表名'
           |    ,info_risk_level STRING COMMENT '变更风险等级'
           |    ,winhc_suggest STRING COMMENT '提示信息'
           |    ,create_time STRING COMMENT '创建时间'
           |)
           |COMMENT '企业动态输出表'
           |PARTITIONED BY (ds STRING COMMENT '分区',tn STRING COMMENT '表名')
           |LIFECYCLE 30
           |""".stripMargin)
    }

    //表名（不加前后辍）
    def calc(tableName: String
             , bName: Int = 0 //是否补充cname字段
            ): Unit = {
      val handle = getClazz[CompanyDynamicHandle](s"com.winhc.bigdata.spark.jobs.dynamic.tables.$tableName")

      val types = handle.org_type()
      val colsExclusiveSome = spark.table(s"${project}.ads_change_extract").columns.filter(s => {
        !s.equals("cid") && !s.equals("data") && !s.equals("old_data") && !s.equals("ds") && !s.equals("tn")
      }).seq
      val rdd = sql(
        bName match {
          //默认:无需补全cname字段
          case 0 =>
            s"""
               |SELECT  *,null AS cname
               |FROM    ${project}.ads_change_extract
               |WHERE   ds = '$ds'
               |AND     tn = '$tableName'
               |AND     TYPE in (${types.map("'" + _ + "'").mkString(",")})
               |""".stripMargin
          //需根据cid补全cname字段数据
          case 1 =>
            s"""
               |SELECT A.*,B.cname AS cname
               |FROM(
               |  SELECT  *
               |  FROM    ${project}.ads_change_extract
               |  WHERE   ds = '$ds'
               |  AND     tn = '$tableName'
               |  AND     TYPE in (${types.map("'" + _ + "'").mkString(",")})
               |) AS A
               |LEFT JOIN (
               |    SELECT cid,cname FROM  $project.base_company_mapping
               |    WHERE ds = '${getLastPartitionsOrElse(project + ".base_company_mapping", "0")}'
               |) AS B
               |ON A.cid = B.cid
               |""".stripMargin
          //既需根据cid补全cname字段数据；还需在data字段中根据其中的cid添加对应cname数据
          case 2 =>
            s"""
               |SELECT cid, ${colsExclusiveSome.mkString(",")},old_data, cname, STR_TO_MAP(regexp_replace(concat_ws(',',data),'[{"}]',''),',',':') AS data
               |FROM(
               |  SELECT new_cid AS cid, ${colsExclusiveSome.mkString(",")},to_json(old_data) AS old_data, COLLECT_SET(cname)[0] AS cname, COLLECT_SET(to_json(data)) AS data
               |  FROM(
               |    SELECT new_cid, ${colsExclusiveSome.map("A." + _).mkString(",")}, A.cid, IF(A.data IS NULL, STR_TO_MAP(CONCAT('cname:',IF(cname IS NULL, '',cname)),',',':'), A.data) AS data, old_data, IF(A.new_cid=A.cid,B.cname,null) AS cname
               |    FROM(
               |      SELECT cid AS new_cid, cid, ${colsExclusiveSome.mkString(",")},data, old_data
               |      FROM(
               |        SELECT  *
               |        FROM    ${project}.ads_change_extract
               |        WHERE   ds = '$ds'
               |        AND     tn = '$tableName'
               |        AND     TYPE in (${types.map("'" + _ + "'").mkString(",")})
               |        )
               |      UNION ALL
               |      SELECT cid AS new_cid, data['cid'] AS cid, ${colsExclusiveSome.mkString(",")},null AS data, old_data
               |      FROM(
               |        SELECT  *
               |        FROM    ${project}.ads_change_extract
               |        WHERE   ds = '$ds'
               |        AND     tn = '$tableName'
               |        AND     TYPE in (${types.map("'" + _ + "'").mkString(",")})
               |        )
               |    ) AS A
               |    LEFT JOIN (
               |        SELECT cid,cname FROM  $project.base_company_mapping
               |        WHERE ds = '${getLastPartitionsOrElse(project + ".base_company_mapping", "0")}'
               |    ) AS B
               |    ON A.cid = B.cid
               |  )
               |  GROUP BY new_cid, ${colsExclusiveSome.mkString(",")},to_json(old_data)
               |)
               |""".stripMargin
        })
        .rdd.flatMap(r => {
        val rowkey = r.getAs[String]("rowkey")
        val cid = r.getAs[String]("cid")
        val new_data = r.getAs[Map[String, String]]("data")
        val old_data = r.getAs[Map[String, String]]("old_data")
        val biz_date = r.getAs[String]("biz_date")
        val fields = r.getAs[String]("fields")
        val cname = r.getAs[String]("cname")
        val result = handle.handle(rowkey, biz_date, cid, if (fields == null) null else fields.split(","), old_data, new_data, cname)
        if (result == null) {
          None
        }
        else {
          result.map(res => Row(res._1, res._2, res._3, res._4, res._5, res._6, res._7, res._8, res._9, res._10, DateFormatUtils.format(new Date(), "yyyy-MM-dd HH:mm:ss")))
        }
      })

      val schema = getSchema(ListMap(
        "cid" -> StringType
        , "cname" -> StringType
        , "info_type" -> StringType
        , "rta_desc" -> StringType
        , "change_content" -> StringType
        , "change_time" -> StringType
        , "biz_id" -> StringType
        , "sub_info_type" -> StringType
        , "info_risk_level" -> StringType
        , "winhc_suggest" -> StringType
        , "create_time" -> StringType
      ))
      spark.createDataFrame(rdd, schema)
        .createOrReplaceTempView("company_dynamic_tmp")

      val cols = getColumns(s"$project.$targetTab").filter(!_.equals("ds")).filter(!_.equals("tn"))

      sql(
        s"""
           |INSERT ${if (isWindows) "INTO" else "OVERWRITE"} TABLE ${getEnvProjectName(env, project)}.$targetTab PARTITION(ds='$ds',tn='$tableName')
           |SELECT ${cols.mkString(",")}
           |FROM
           |    company_dynamic_tmp
           |""".stripMargin)
    }
  }

  private val startArgs = Seq(
    Args(tableName = "company_abnormal_info", bName = 0)
    , Args(tableName = "company_equity_info")
  )

  private case class Args(project: String = "winhc_eci_dev"
                          , tableName: String
                          , bName: Int = 1)

  def main(args: Array[String]): Unit = {


    if (args.length != 3) {
      println(
        s"""
           |Please enter the legal parameters !
           |<project> <ds> <tableNames>
           |""".stripMargin)
      sys.exit(-99)
    }

    val Array(project, tableNames, ds) = args

    println(
      s"""
         |project: $project
         |tableNames: $tableNames
         |ds: $ds
         |""".stripMargin)

    val config = EsConfig.getEsConfigMap ++ mutable.Map(
      "spark.hadoop.odps.project.name" -> project,
      "spark.hadoop.odps.spark.local.partition.amt" -> "10"
    )
    val spark = SparkUtils.InitEnv("CompanyDynamic", config)
    val cd = CompanyDynamicUtil(spark, project, ds)
    cd.init()

    val ts = tableNames.split(",").toSet

    startArgs.filter(e => {
      ts.contains(e.tableName)
    }).foreach(e => {
      cd.calc(e.tableName, e.bName)
    })
    spark.stop()
  }
}
